/* 
 * Metronome session
 *
 * This file is part of GTick
 * 
 *
 * Copyright (c) 1999, Alex Roberts
 * Copyright (c) 2003, Roland Stigge <stigge@antcom.de>
 *
 * GTick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GTick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GTick; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 
#include <config.h>

/* GNU headers */
#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <math.h>

/* GTK+ headers */
#include <gtk/gtk.h>
#include <glib.h>

/* own headers */
#include "globals.h"
#include "metro.h"
#include "dsp.h"
#include "help.h"
#include "gtkutil.h"
#include "manualtimer.h"
#include "util.h"
#include "option.h"
#include "gtkoptions.h"
#include "threadtalk.h"

/* icons */
#include "icon32x32.xpm"
#include "icon48x48.xpm"
#include "icon64x64.xpm"

#define TIMER_DELAY 300
volatile int interrupted = 0; /* the caught signal will be stored here */

static state_data_t state_data[] = {
  { N_("Idle"),    N_("Start") },  /* STATE_IDLE    */
  { N_("Running"), N_("Stop") }    /* STATE_RUNNING */
};

/* handler for SIGTERM and SIGINT */
static void terminate_signal_callback(int sig) {
  interrupted = sig;
}

/*
 * read meter from GUI (widgets)
 */
static int gui_get_meter(metro_t* metro) {
  int meter;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(metro->meter_button_1)))
    meter = 1;
  else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(
	  metro->meter_button_2)))
    meter = 2;
  else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(
	  metro->meter_button_3)))
    meter = 3;
  else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(
	  metro->meter_button_4)))
    meter = 4;
  else
    meter = gtk_spin_button_get_value_as_int(
	GTK_SPIN_BUTTON(metro->meter_spin_button));

  return meter;
}

/*
 * GUI volume change callback
 */
static void set_volume_cb(metro_t* metro)
{
  int* volume = (int*) malloc(sizeof(int));

  if (volume == NULL) {
    printf("Error: Insufficient memory for allocating volume message.\n");
    exit(1);
  }
  *volume = GTK_ADJUSTMENT(metro->volume_adjustment)->value;
  comm_client_query(metro->inter_thread_comm, MESSAGE_TYPE_SET_VOLUME, volume);
}

/*
 * GUI rate change callback
 */
static void set_speed_cb(metro_t *metro)
{
  double* frequency = (double*) malloc(sizeof(double));
  
  if (frequency == NULL) {
    printf("Error: Insufficient memory for allocating frequency message.\n");
    exit(1);
  }
  *frequency = GTK_ADJUSTMENT(metro->speed_adjustment)->value / 60.0;
  if (debug)
    g_print ("set_speed_cb: rate=%f bpm\n", *frequency * 60.0);
  comm_client_query(metro->inter_thread_comm,
                    MESSAGE_TYPE_SET_FREQUENCY,
		    frequency);
}

/*
 * Callback for meter mode
 */
static void set_meter_cb(metro_t *metro) {
  int* meter = (int*) malloc (sizeof(int));

  if (meter == NULL) {
    printf("Error: Insufficient memory for allocating meter message.\n");
    exit(1);
  }
  
  *meter = gui_get_meter(metro);
  gtk_widget_set_sensitive(metro->meter_spin_button,
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(metro->meter_button_more)));
  
  if (debug)
    g_print("new meter: %d\n", *meter);

  comm_client_query(metro->inter_thread_comm, MESSAGE_TYPE_SET_METER, meter);
}

/*
 * Change state
 */
static void set_state(metro_t* metro, int state) {
 
  if (state != metro->state) {
    
    if (metro->state == STATE_RUNNING && state == STATE_IDLE) {
      comm_client_query(metro->inter_thread_comm,
	                MESSAGE_TYPE_STOP_METRONOME, NULL);
      execute(metro->options->command_on_stop);
    } else if (metro->state == STATE_IDLE && state == STATE_RUNNING) {
      comm_client_query(metro->inter_thread_comm,
	                MESSAGE_TYPE_START_METRONOME, NULL);
      execute(metro->options->command_on_start);
    } else {
      fprintf(stderr, "Warning: Unhandled state change.\n");
    }
    
    gtk_statusbar_pop(GTK_STATUSBAR(metro->statusbar),
		      metro->status_context);
    gtk_statusbar_push(GTK_STATUSBAR(metro->statusbar),
		       metro->status_context,
		       _(state_data[state].state));
    gtk_label_set_text(GTK_LABEL(metro->togglebutton_label),
		       _(state_data[state].toggle_label));
    
    metro->state = state;
  }
}

/*
 * handle messages from server
 */
static void handle_comm(metro_t* metro) {
  message_type_t message_type;
  void* body = NULL;
  int newvolume = -1;

  while ((message_type = comm_client_try_get_reply(metro->inter_thread_comm,
	                                           &body)) !=
         MESSAGE_TYPE_NO_MESSAGE)
  {
    switch (message_type) {
      case MESSAGE_TYPE_RESPONSE_VOLUME:
	newvolume = *((int*) body);
	free(body);
	break;
      case MESSAGE_TYPE_RESPONSE_START_ERROR:
	gtk_widget_show(metro->start_error);
	
	set_state(metro, STATE_IDLE);
        break;
      default:
        printf("Warning: Unhandled message type: %d.\n", message_type);
    }
  }

  if (newvolume >= 0) {
    gtk_adjustment_set_value(GTK_ADJUSTMENT(metro->volume_adjustment),
			     newvolume);
    if (!metro->volume_initialized) {
      g_signal_connect_swapped(G_OBJECT(metro->volume_adjustment),
			       "value_changed",
			       G_CALLBACK(set_volume_cb),
			       metro);
      metro->volume_initialized = 1;
    }
  }
}

/*
 * Callback for toggle button (metronome on/off ...)
 */
static void metro_toggle_cb(metro_t* metro) {
  int new_state;
  
  if (metro->state == STATE_IDLE) {
    new_state = STATE_RUNNING;
  } else {
    new_state = STATE_IDLE;
  }

  set_state(metro, new_state);
}

/*
 * Quit program and destroy callback
 */
static void quit_cb(metro_t *metro, guint action _U_, GtkWidget* widget _U_)
{
  set_state(metro, STATE_IDLE);
  option_save_all(metro->options->option_list);
  comm_client_query(metro->inter_thread_comm, MESSAGE_TYPE_STOP_SERVER, NULL);
  gtk_main_quit();
}

/*
 * periodically (e.g. each 300 milliseconds) from gtk main loop called function
 */
static gint timeout_callback(metro_t* metro) {
  if (interrupted) {
    switch(interrupted) {
    case SIGINT:
      if (debug)
	fprintf(stderr, "SIGINT (Ctrl-C ?) caught.\n");
      break;
    case SIGTERM:
      if (debug)
	fprintf(stderr, "SIGTERM caught.\n");
      break;
    default:
      fprintf(stderr, "Warning: Unknown signal caught.\n");
    }
    quit_cb(metro, 0, NULL);
  }

  comm_client_query(metro->inter_thread_comm,
                    MESSAGE_TYPE_GET_VOLUME_CHANGE, NULL);
  handle_comm(metro);

  return TRUE; /* call it again */
}

/*
 * indirecting callback showing tapping_dialog_new()
 */
static void manual_timer_cb(metro_t* metro,
                          guint action _U_,
			  GtkWidget* widget _U_)
{
  GtkWidget* dialog = tapping_dialog_new(metro->window,
                                         metro->speed_adjustment);

  gtk_widget_show(dialog);
}

/*
 * indirecting callback showing gtk_options_dialog_new()
 */
static void properties_cb(metro_t* metro,
                         guint action _U_,
			 GtkWidget* widget _U_)
{
  GtkWidget* dialog = gtk_options_dialog_new(metro->window, metro->options);

  gtk_widget_show(dialog);
}

/*
 * option system callback for initializing sample name with default value
 * returns 0 on success, -1 otherwise
 */
static int new_sample(metro_t* metro) {
  metro->options->sample_name = strdup(DEFAULT_SAMPLE_FILENAME);
  comm_client_query(metro->inter_thread_comm,
                    MESSAGE_TYPE_SET_SOUND, strdup(DEFAULT_SAMPLE_FILENAME));
  if (metro->options->sample_name)
    return 0;
  else
    return -1;
}

/*
 * option system callback for destroying sample name
 */
static void delete_sample(metro_t* metro) {
  if (metro->options->sample_name)
    free(metro->options->sample_name);
}

/*
 * option system callback for choosing the sample
 *
 * returns 0 on success, -1 otherwise
 */
static int set_sample(metro_t* metro, const char* sample_name) {
  if (metro->options->sample_name)
    free(metro->options->sample_name);
  else
    fprintf(stderr,
	    "free() error: metro->options->sample_name not allocated.\n");
  metro->options->sample_name = strdup(sample_name);
  comm_client_query(metro->inter_thread_comm,
                    MESSAGE_TYPE_SET_SOUND, strdup(sample_name));
  if (metro->state == STATE_RUNNING) {
    comm_client_query(metro->inter_thread_comm,
	              MESSAGE_TYPE_STOP_METRONOME, NULL);
    comm_client_query(metro->inter_thread_comm,
	              MESSAGE_TYPE_START_METRONOME, NULL);
  }

  return 0;
}

/*
 * option system callback for spotting the sample
 */
static const char* get_sample(metro_t* metro) {
  return metro->options->sample_name;
}

/*
 * option system callback for initializing sound device name with default value
 * returns 0 on success, -1 otherwise
 */
static int new_sound_device_name(metro_t* metro) {
  metro->options->sound_device_name = strdup(DEFAULT_SOUND_DEVICE_FILENAME);
  comm_client_query(metro->inter_thread_comm,
                    MESSAGE_TYPE_SET_DEVICE,
		    strdup(DEFAULT_SOUND_DEVICE_FILENAME));
  if (metro->options->sound_device_name)
    return 0;
  else
    return -1;
}

/*
 * option system callback for destroying sound device name
 */
static void delete_sound_device_name(metro_t* metro) {
  if (metro->options->sound_device_name)
    free(metro->options->sound_device_name);
}

/*
 * option system callback for choosing the sound device name
 *
 * returns 0 on success, -1 otherwise
 */
static int set_sound_device_name(metro_t* metro, const char* sound_device_name)
{
  if (metro->options->sound_device_name)
    free(metro->options->sound_device_name);
  else
    fprintf(stderr,
	    "free() error: metro->options->sound_device_name not allocated.\n");
  metro->options->sound_device_name = strdup(sound_device_name);
  comm_client_query(metro->inter_thread_comm,
                    MESSAGE_TYPE_SET_DEVICE, strdup(sound_device_name));
  if (metro->state == STATE_RUNNING) {
    comm_client_query(metro->inter_thread_comm,
	              MESSAGE_TYPE_STOP_METRONOME, NULL);
    comm_client_query(metro->inter_thread_comm,
	              MESSAGE_TYPE_START_METRONOME, NULL);
  }

  return 0;
}

/*
 * option system callback for spotting the sound device name
 */
static const char* get_sound_device_name(metro_t* metro) {
  return metro->options->sound_device_name;
}

/*
 * option system callback for choosing the speed
 *
 * returns 0 on success, -1 otherwise
 */
static int set_speed(metro_t* metro, const char* speed)
{
  char* saved_locale;

  if (!metro || !metro->speed_adjustment || !speed)
    return -1;
  
  saved_locale = strdup (setlocale (LC_NUMERIC, NULL));
  if (saved_locale == NULL) {
    fprintf(stderr, "set_speed: Out of memory.\n");
    exit(1);
  }
  setlocale (LC_NUMERIC, "C");

  gtk_adjustment_set_value(GTK_ADJUSTMENT(metro->speed_adjustment),
                           strtod(speed, NULL));
  
  setlocale (LC_NUMERIC, saved_locale);
  free(saved_locale);

  return 0;
}

/*
 * option system callback for spotting the speed
 *
 * if called with metro == NULL, deinitializes state and return NULL
 */
static const char* get_speed(metro_t* metro) {
  static char* result = NULL;

  if (result)
    free(result);
  
  if (metro == NULL) {
    return NULL;
  } else {
    char* saved_locale;

    saved_locale = strdup (setlocale (LC_NUMERIC, NULL));
    if (saved_locale == NULL) {
      fprintf(stderr, "set_speed: Out of memory.\n");
      exit(1);
    }
    setlocale (LC_NUMERIC, "C");

    if (asprintf(&result, "%.1f",
          gtk_adjustment_get_value(GTK_ADJUSTMENT(metro->speed_adjustment)))
	== -1)
    {
      fprintf(stderr,
	      "get_speed: Couldn't allocate memory for result string.\n");
      exit(1);
    }
  
    setlocale (LC_NUMERIC, saved_locale);
    free(saved_locale);
    
    return result;
  }
}

/*
 * option system callback for initializing speed option
 */
static int new_speed(metro_t* metro) {
  if (metro->speed_adjustment) {
    gtk_adjustment_set_value(GTK_ADJUSTMENT(metro->speed_adjustment),
                             DEFAULT_SPEED);
    return 0;
  } else {
    return -1;
  }
}

/*
 * option system callback for destroying speed option
 */
static void delete_speed(metro_t* metro _U_) {
  get_speed(NULL);
}

/*
 * option system callback for choosing the meter
 *
 * returns 0 on success, -1 otherwise
 */
static int set_meter(metro_t* metro, const char* meter)
{
  int result;
  
  if (!metro || !meter)
    return -1;

  result = (int) strtol(meter, NULL, 0);

  if (result != gui_get_meter(metro)) {
    switch (result) {
      case 1:
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(metro->meter_button_1),
	    TRUE);
	break;
      case 2:
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(metro->meter_button_2),
	    TRUE);
	break;
      case 3:
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(metro->meter_button_3),
	    TRUE);
        break;
      case 4:
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(metro->meter_button_4),
	    TRUE);
	break;
      default:
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(metro->meter_button_more), TRUE);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(metro->meter_spin_button),
	                          result);
    }
    gtk_widget_set_sensitive(metro->meter_spin_button, result > 4);
  }
  return 0;
}

/*
 * option system callback for spotting the meter
 *
 * if called with metro == NULL, deinitializes state and return NULL
 */
static const char* get_meter(metro_t* metro) {
  static char* result = NULL;

  if (result)
    free(result);
  
  if (metro == NULL) {
    return NULL;
  } else {
    if (asprintf(&result, "%d", gui_get_meter(metro)) == -1) {
      fprintf(stderr,
	      "get_meter: Couldn't allocate memory for result string.\n");
      exit(1);
    }
  
    return result;
  }
}

/*
 * option system callback for initializing meter option
 */
static int new_meter(metro_t* metro) {
  char* s;
  int* meter;

  if (asprintf(&s, "%d", DEFAULT_METER) < 0)
    return -1;
  set_meter(metro, s);
  free(s);
  
  if ((meter = (int*) malloc (sizeof(int))) == NULL)
    return -1;
  *meter = DEFAULT_METER;
  comm_client_query(metro->inter_thread_comm, MESSAGE_TYPE_SET_METER, meter);

  return 0;
}

/*
 * option system callback for destroying meter option
 */
static void delete_meter(metro_t* metro _U_) {
  get_meter(NULL);
}

/* option system callback for initializing command for metronome start */
static int new_command_on_start(options_t* options) {
  if ((options->command_on_start = strdup(DEFAULT_COMMAND_ON_START)))
    return 0;
  else
    return -1;
}

/* option system callback for destroying command for metronome start */
static void delete_command_on_start(options_t* options) {
  if (options->command_on_start)
    free(options->command_on_start);
}

/* option system callback for setting command for metronome start */
static int set_command_on_start(options_t* options, const char* command) {
  if (options->command_on_start)
    free(options->command_on_start);
  if ((options->command_on_start = strdup(command)))
    return 0;
  else
    return -1;
}

/* option system callback for getting command for metronome start */
static const char* get_command_on_start(options_t* options) {
  return options->command_on_start;
}

/* option system callback for initializing command for metronome stop */
static int new_command_on_stop(options_t* options) {
  if ((options->command_on_stop = strdup(DEFAULT_COMMAND_ON_STOP)))
    return 0;
  else
    return -1;
}

/* option system callback for destroying command for metronome stop */
static void delete_command_on_stop(options_t* options) {
  if (options->command_on_stop)
    free(options->command_on_stop);
}

/* option system callback for setting command for metronome stop */
static int set_command_on_stop(options_t* options, const char* command) {
  if (options->command_on_stop)
    free(options->command_on_stop);
  if ((options->command_on_stop = strdup(command)))
    return 0;
  else
    return -1;
}

/* option system callback for getting command for metronome stop */
static const char* get_command_on_stop(options_t* options) {
  return options->command_on_stop;
}

/*
 * returns the (main) menu bar for the specified window
 */
static GtkWidget* get_menubar(metro_t* metro) {
  GtkItemFactoryEntry menu_items[] = {
  /* path, accelerator, cb, cb parameter, item type, extra */
{_("/_Metronome"), NULL, NULL, 0, "<Branch>", NULL},
{_("/Metronome/_Manual Timer"), NULL, manual_timer_cb, 1, NULL, NULL},
{_("/Metronome/"), NULL, NULL, 0, "<Separator>", NULL},
{_("/Metronome/_Quit"), "<control>Q", quit_cb, 1, "<StockItem>", GTK_STOCK_QUIT},

{_("/_Options"), NULL, NULL, 0, "<Branch>", NULL},
{_("/Options/_Properties"), NULL, properties_cb, 1, "<StockItem>", GTK_STOCK_PROPERTIES},

{_("/_Help"), NULL, NULL, 0, "<LastBranch>", NULL},
{_("/Help/_About"), NULL, about_box, 0, NULL, NULL},
{_("/Help/_License"), NULL, license_box, 0, NULL, NULL}
  };
  gint nmenu_items = sizeof(menu_items) / sizeof(menu_items[0]);
  GtkAccelGroup* accel_group = gtk_accel_group_new();
  GtkItemFactory* item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR,
                                                      "<main>", accel_group);

  gtk_item_factory_create_items(item_factory, nmenu_items, menu_items, metro);
  gtk_window_add_accel_group(GTK_WINDOW(metro->window), accel_group);

  return gtk_item_factory_get_widget(item_factory, "<main>");
}

/*
 * The audio thread
 */
static gpointer audio_loop(metro_t* metro) {
  dsp_t* dsp;

  dsp = dsp_new(metro->inter_thread_comm);
  dsp_main_loop(dsp);
  dsp_delete(dsp);

  return NULL;
}

/*
 * returns a new metro object
 */
metro_t* metro_new(void) {
  metro_t* metro;
  GtkObject* adjustment;
  GList* icon_list = NULL;
  GtkWidget* windowvbox;
  GtkWidget* vbox;
  GtkWidget* metroframe;
  GtkWidget* metrovbox;
  GtkWidget* menu;
  GtkWidget* label;
  GtkWidget* table;
  GtkWidget* beatbox;
  GtkWidget* separator;
  GtkWidget* spinbutton;
  GtkWidget* hbox;
  GtkWidget* togglebutton;
  GtkWidget* scale;

  if (!(metro = calloc(1, sizeof(metro_t)))) {
    fprintf(stderr, "Error allocating memory for metro object.\n");
    exit(1);
  }
  if (debug)
    printf("Creating metro object at %p.\n", metro);
  
  metro->options = options_new();
  metro->state = STATE_IDLE; /* set default */
  
  metro->inter_thread_comm = comm_new();
  metro->audio_thread =
    g_thread_create_full((GThreadFunc) audio_loop,
                         metro, /* data passed to function */
		         0, /* stack size = default */
			 TRUE, /* joinable */
			 TRUE, /* bound in system scope (faster) */
			 G_THREAD_PRIORITY_NORMAL, /* G_THREAD_PRIORITY_URGENT
						   would be highest priority */
			 NULL /* no GError reporting */);
  
  option_register(&metro->options->option_list,
                  "SampleFilename",
		  (option_new_t) new_sample,
		  (option_delete_t) delete_sample,
		  (option_set_t) set_sample,
		  (option_get_t) get_sample,
		  (void*) metro);
  option_register(&metro->options->option_list,
                  "SoundDevice",
		  (option_new_t) new_sound_device_name,
		  (option_delete_t) delete_sound_device_name,
		  (option_set_t) set_sound_device_name,
		  (option_get_t) get_sound_device_name,
		  (void*) metro);

  option_register(&metro->options->option_list,
                  "CommandOnStart",
		  (option_new_t) new_command_on_start,
		  (option_delete_t) delete_command_on_start,
		  (option_set_t) set_command_on_start,
		  (option_get_t) get_command_on_start,
		  (void*) metro->options);
  option_register(&metro->options->option_list,
                  "CommandOnStop",
		  (option_new_t) new_command_on_stop,
		  (option_delete_t) delete_command_on_stop,
		  (option_set_t) set_command_on_stop,
		  (option_get_t) get_command_on_stop,
		  (void*) metro->options);
  
  metro->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  g_signal_connect_swapped(G_OBJECT(metro->window),
                           "destroy",
		           G_CALLBACK(quit_cb),
		           metro);

  gtk_window_set_title (GTK_WINDOW (metro->window), TITLE " " VERSION);

  icon_list = g_list_append(icon_list,
          (gpointer) gdk_pixbuf_new_from_xpm_data((const char**) icon32x32));
  icon_list = g_list_append(icon_list,
          (gpointer) gdk_pixbuf_new_from_xpm_data((const char**) icon48x48));
  icon_list = g_list_append(icon_list,
          (gpointer) gdk_pixbuf_new_from_xpm_data((const char**) icon64x64));
  gtk_window_set_icon_list(GTK_WINDOW(metro->window), icon_list);

  windowvbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add(GTK_CONTAINER(metro->window), windowvbox);
  gtk_widget_show (windowvbox);

  menu = get_menubar(metro);
  gtk_box_pack_start(GTK_BOX(windowvbox), menu, FALSE, TRUE, 0);
  gtk_widget_show(menu);

  vbox = gtk_vbox_new(FALSE, 8);
  gtk_box_pack_start_defaults(GTK_BOX(windowvbox), vbox);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 8);
  gtk_widget_show(vbox);

  metro->statusbar = gtk_statusbar_new();
  metro->status_context =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(metro->statusbar), "status");
  gtk_statusbar_push(GTK_STATUSBAR(metro->statusbar),
                     metro->status_context, 
		     _(state_data[metro->state].state));
  gtk_box_pack_end(GTK_BOX(windowvbox), metro->statusbar, FALSE, TRUE, 0);
  gtk_widget_show(metro->statusbar);

  /* Metronome UI Code */
  metroframe = gtk_frame_new(_("Settings"));
  gtk_box_pack_start_defaults(GTK_BOX(vbox), metroframe);
  gtk_widget_show (metroframe);

  metrovbox = gtk_vbox_new(FALSE, 10);
  gtk_container_set_border_width(GTK_CONTAINER(metrovbox), 10);
  gtk_container_add(GTK_CONTAINER(metroframe), metrovbox);
  gtk_widget_show(metrovbox);

  table = gtk_table_new(2, 3, FALSE);
  gtk_box_pack_start_defaults(GTK_BOX(metrovbox), table);
  gtk_table_set_row_spacings(GTK_TABLE(table), 10);
  gtk_table_set_col_spacings(GTK_TABLE(table), 10);
  gtk_widget_show(table);
  
  label = gtk_label_new(_("Speed:"));
  gtk_table_attach(GTK_TABLE(table), label,
                   0, 1, 0, 1,
		   GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
  gtk_widget_show (label);

  metro->speed_adjustment = gtk_adjustment_new(100.0, MIN_BPM, MAX_BPM,
                                         1.0, 1.0, 0.0);
  g_signal_connect_swapped(G_OBJECT(metro->speed_adjustment),
                           "value_changed",
		           G_CALLBACK(set_speed_cb),
		           metro);
  set_speed_cb(metro);

  scale = gtk_hscale_new (GTK_ADJUSTMENT (metro->speed_adjustment));
  gtk_range_set_update_policy (GTK_RANGE (scale), GTK_UPDATE_CONTINUOUS);
  gtk_scale_set_digits (GTK_SCALE (scale), 1);
  gtk_scale_set_draw_value (GTK_SCALE (scale), FALSE);
  gtk_table_attach(GTK_TABLE(table), scale,
                   1, 2, 0, 1,
		   GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_widget_show (scale);

  spinbutton = gtk_spin_button_new(GTK_ADJUSTMENT(metro->speed_adjustment),
                                   1.0, /* Step */
				   1    /* number of decimals */);
  gtk_table_attach(GTK_TABLE(table), spinbutton,
                   2, 3, 0, 1,
		   GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_widget_show(spinbutton);

  option_register(&metro->options->option_list,
                  "Speed",
		  (option_new_t) new_speed,
		  (option_delete_t) delete_speed,
		  (option_set_t) set_speed,
		  (option_get_t) get_speed,
		  (void*) metro);
  
  label = gtk_label_new (_("Volume:"));
  gtk_table_attach(GTK_TABLE(table), label,
                   0, 1, 1, 2,
		   GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
  gtk_widget_show (label);
	  
  metro->volume_adjustment = gtk_adjustment_new (0.0, VOLUME_MIN, VOLUME_MAX,
                                    5.0, 25.0, 0.0);
  /* we set up the callback later, only when the mixer sent an initial volume */
  
  scale = gtk_hscale_new (GTK_ADJUSTMENT(metro->volume_adjustment));
  gtk_range_set_update_policy (GTK_RANGE (scale), GTK_UPDATE_CONTINUOUS);
  gtk_scale_set_digits (GTK_SCALE (scale), 0);
  gtk_scale_set_draw_value (GTK_SCALE (scale), FALSE);
  gtk_table_attach(GTK_TABLE(table), scale,
                   1, 2, 1, 2,
		   GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_widget_show (scale);
  
  spinbutton = gtk_spin_button_new(GTK_ADJUSTMENT(metro->volume_adjustment),
                                   1.0, /* Step */
				   1    /* number of decimals */);
  gtk_table_attach(GTK_TABLE(table), spinbutton,
                   2, 3, 1, 2,
		   GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_widget_show(spinbutton);
  
  /* Timing UI Code */
  
  beatbox = gtk_frame_new (_("Beat"));
  gtk_box_pack_start_defaults(GTK_BOX(metrovbox), beatbox);
  gtk_widget_show (beatbox);

  hbox = gtk_hbox_new (FALSE, 20);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 10);
  gtk_container_add (GTK_CONTAINER (beatbox), hbox);
  gtk_widget_show (hbox);

  metro->meter_button_1 = gtk_radio_button_new_with_label (NULL,
      /* TRANSLATORS: all beats are equal */
      _("Even"));
  gtk_box_pack_start (GTK_BOX (hbox), metro->meter_button_1, TRUE, TRUE, 0);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (metro->meter_button_1), TRUE);
  gtk_widget_show (metro->meter_button_1);

  g_signal_connect_swapped(G_OBJECT(metro->meter_button_1),
                           "toggled",
		           G_CALLBACK(set_meter_cb),
		           metro);
  
  metro->meter_button_2 =
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(metro->meter_button_1),
	/* TRANSLATORS: duple meter */
	_("2/4"));
  gtk_box_pack_start(GTK_BOX(hbox), metro->meter_button_2, TRUE, TRUE, 0);
  gtk_widget_show (metro->meter_button_2);
  g_signal_connect_swapped(G_OBJECT(metro->meter_button_2),
                           "toggled",
		           G_CALLBACK(set_meter_cb),
		           metro);
  
  metro->meter_button_3 =
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(metro->meter_button_2),
	/* TRANSLATORS: triple meter */
	_("3/4"));
  gtk_box_pack_start(GTK_BOX(hbox), metro->meter_button_3, TRUE, TRUE, 0);
  gtk_widget_show (metro->meter_button_3);
  g_signal_connect_swapped(G_OBJECT(metro->meter_button_3),
                           "toggled",
		           G_CALLBACK(set_meter_cb),
		           metro);

  metro->meter_button_4 =
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(metro->meter_button_3),
	/* TRANSLATORS: quadruple meter / common meter */
	_("4/4"));
  gtk_box_pack_start(GTK_BOX(hbox), metro->meter_button_4, TRUE, TRUE, 0);
  gtk_widget_show (metro->meter_button_4);
  g_signal_connect_swapped(G_OBJECT(metro->meter_button_4),
                           "toggled",
		           G_CALLBACK(set_meter_cb),
		           metro);

  metro->meter_button_more =
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(metro->meter_button_4),
	/* TRANSLATORS: meter with more than 4 beats per measure */
	_("Other:"));
  gtk_box_pack_start(GTK_BOX(hbox), metro->meter_button_more,
      TRUE, TRUE, 0);
  gtk_widget_show(metro->meter_button_more);
  g_signal_connect_swapped(G_OBJECT(metro->meter_button_more),
                           "toggled",
                           G_CALLBACK(set_meter_cb),
	                   metro);
  
  adjustment = gtk_adjustment_new(5, 1, 99, 1, 4, 0);
  metro->meter_spin_button =
    gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 1, 0);
  gtk_widget_set_sensitive(metro->meter_spin_button, FALSE);
  gtk_widget_show(metro->meter_spin_button);
  gtk_box_pack_start(GTK_BOX(hbox), metro->meter_spin_button,
      TRUE, TRUE, 0);
  g_signal_connect_swapped(G_OBJECT(metro->meter_spin_button),
                           "value-changed",
                           G_CALLBACK(set_meter_cb),
		           metro);

  option_register(&metro->options->option_list,
                  "Meter",
		  (option_new_t) new_meter,
		  (option_delete_t) delete_meter,
		  (option_set_t) set_meter,
		  (option_get_t) get_meter,
		  (void*) metro);
  
  /* General UI Code */
  separator = gtk_hseparator_new ();
  gtk_box_pack_start_defaults(GTK_BOX(windowvbox), separator);
  gtk_widget_show (separator);

  hbox = gtk_hbox_new (FALSE, 10);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 5);
  gtk_box_pack_start (GTK_BOX(windowvbox), hbox, FALSE, TRUE, 0);
  gtk_widget_show (hbox);

  togglebutton = gtk_button_new();
  g_signal_connect_swapped(G_OBJECT(togglebutton),
                           "clicked",
		           G_CALLBACK(metro_toggle_cb),
		           metro);
			   
  gtk_box_pack_start(GTK_BOX(hbox), togglebutton, TRUE, FALSE, 0);
  GTK_WIDGET_SET_FLAGS (togglebutton, GTK_CAN_DEFAULT);
  gtk_widget_grab_default (togglebutton);
  GTK_WIDGET_SET_FLAGS (togglebutton, GTK_CAN_FOCUS);
  gtk_widget_grab_focus (togglebutton);
  gtk_widget_show (togglebutton);

  metro->togglebutton_label =
    gtk_label_new(_(state_data[metro->state].toggle_label));
  gtk_misc_set_padding(GTK_MISC(metro->togglebutton_label), 64, 0);
  gtk_container_add(GTK_CONTAINER(togglebutton),
		    metro->togglebutton_label);
  gtk_widget_show(metro->togglebutton_label);

  /* read rc file */
  option_restore_all(metro->options->option_list);

  comm_client_query(metro->inter_thread_comm, MESSAGE_TYPE_GET_VOLUME, NULL);

  /* set up additional handlers */
  gtk_timeout_add(TIMER_DELAY, (GtkFunction) timeout_callback, metro);
  
  signal(SIGINT, &terminate_signal_callback);
  signal(SIGTERM, &terminate_signal_callback);

  /* additional helper widgets: */
  
  /* Start error message */
  metro->start_error = gtk_message_dialog_new(GTK_WINDOW(metro->window),
            GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
            GTK_MESSAGE_ERROR,
            GTK_BUTTONS_CLOSE,
            _("Couldn't start metronome.\n"
	      "Please check if specified sound device\n"
	      "and sample file are accessible."));
  /* Hide the dialog when the user responds to it */
  g_signal_connect_swapped(G_OBJECT(metro->start_error),
                           "response",
                           G_CALLBACK(gtk_widget_hide),
                           G_OBJECT(metro->start_error));
	
  gtk_widget_show (metro->window);

  return metro;
}

void metro_delete(metro_t* metro) {
  g_thread_join(metro->audio_thread);
  comm_delete(metro->inter_thread_comm);
  metro->inter_thread_comm = NULL;
  options_delete(metro->options);
  if (debug)
    printf("Destroying metro object at %p.\n", metro);
  g_free(metro);
}
